package cmd

import (
	"encoding/xml"
	"fmt"
	"io"
	"net/http"
	"os"
	"path"
	"path/filepath"
	"time"

	"github.com/matthewjohn/archive-org-downloader/archiveorg"
	"github.com/rivo/tview"
)

func Run() {
	archiveDetails := getArchiveDetails()
	files := getArchiveContent(&archiveDetails)
	for _, file := range files.Files {
		println(file.Name)
	}

	checkDestinationDirectory(&archiveDetails)

	userSelectFiles(files)

	if !finalConfirm() {
		return
	}

	downloadFiles(&archiveDetails, files)
	println("All files downloaded")
}

type ProgressWriter struct {
	Total       int64
	LastUpdate  time.Time
	UpdateFunc  func(int64)
	ContentSize int64
}

func (pw *ProgressWriter) Write(p []byte) (int, error) {
	n := len(p)
	pw.Total += int64(n)

	// Only update periodically to avoid spamming UI
	if time.Since(pw.LastUpdate) > 200*time.Millisecond {
		pw.UpdateFunc(pw.Total)
		pw.LastUpdate = time.Now()
	}
	return n, nil
}

func downloadFiles(archiveDetails *archiveorg.ArchiveDetails, archiveContent *archiveorg.ArchiveContent) {
	for _, file := range archiveContent.Files {
		if file.Selected {
			downloadFile(archiveDetails, &file)
		}
	}
}

func downloadFile(archiveDetails *archiveorg.ArchiveDetails, archiveFile *archiveorg.ArchiveFile) {
	app := tview.NewApplication()
	textView := tview.NewTextView().
		SetDynamicColors(true).
		SetChangedFunc(func() {
			app.Draw()
		})

	textView.SetText("Preparing download...\n")

	// Layout
	flex := tview.NewFlex().SetDirection(tview.FlexRow).
		AddItem(textView, 0, 1, true)

	// Start download in background
	go func() {
		resp, err := archiveDetails.DownloadFile(archiveFile.Name)
		if err != nil {
			app.QueueUpdateDraw(func() {
				textView.Write([]byte(fmt.Sprintf("[red]Failed to download: %v\n", err)))
			})
			return
		}
		defer resp.Body.Close()

		if resp.StatusCode != http.StatusOK {
			app.QueueUpdateDraw(func() {
				textView.Write([]byte(fmt.Sprintf("[red]HTTP error: %s\n", resp.Status)))
			})
			return
		}

		destination := path.Join(archiveDetails.DestinationDirectory, archiveFile.Name)

		err = os.MkdirAll(filepath.Dir(destination), 755)
		if err != nil {
			app.QueueUpdateDraw(func() {
				textView.Write([]byte(fmt.Sprintf("[red]Failed to create directory: %v\n", err)))
			})
			return
		}

		file, err := os.Create(destination)
		if err != nil {
			app.QueueUpdateDraw(func() {
				textView.Write([]byte(fmt.Sprintf("[red]Failed to create file: %v\n", err)))
			})
			return
		}
		defer file.Close()

		contentLength := resp.ContentLength

		progress := &ProgressWriter{
			UpdateFunc: func(downloaded int64) {
				app.QueueUpdateDraw(func() {
					percent := float64(downloaded) / float64(contentLength) * 100
					textView.SetText(fmt.Sprintf("Downloading %s... %.2f%% (%d / %d bytes)", archiveFile.Name, percent, downloaded, contentLength))
				})
			},
			ContentSize: contentLength,
		}

		_, err = io.Copy(io.MultiWriter(file, progress), resp.Body)
		if err != nil {
			app.QueueUpdateDraw(func() {
				textView.Write([]byte(fmt.Sprintf("\n[red]Download error: %v\n", err)))
			})
			return
		}

		app.Stop()
	}()

	if err := app.SetRoot(flex, true).Run(); err != nil {
		panic(err)
	}
}

func finalConfirm() bool {
	app := tview.NewApplication()
	var confirm bool
	list := tview.NewList().AddItem("Continue", "Press to start download", 'c', func() {
		app.Stop()
		confirm = true
	}).AddItem("Exit", "Press to abort", 'q', func() {
		app.Stop()
		confirm = false
	})
	if err := app.SetRoot(list, true).EnableMouse(true).Run(); err != nil {
		panic(err)
	}
	return confirm
}

func checkDestinationDirectory(archiveDetails *archiveorg.ArchiveDetails) {
	file, err := os.Open(archiveDetails.DestinationDirectory)
	if err != nil {
		panic(err)
	}

	defer file.Close()

	fileInfo, err := file.Stat()
	if err != nil {
		panic(err)
	}

	if !fileInfo.IsDir() {
		// file is not a directory
		panic(fmt.Errorf("destination is not a directory"))
	}
}

func userSelectFiles(content *archiveorg.ArchiveContent) {
	// Display all files to user to let them select
	app := tview.NewApplication()
	list := tview.NewList().AddItem("Continue", "Press to download", 'c', func() {
		app.Stop()
	})

	const preItems = 3

	list = list.AddItem("Include all", "", 'a', func() {
		for itx, _ := range content.Files {
			content.Files[itx].Selected = true
			list.SetItemText(itx+preItems, content.Files[itx].Name, "To download")
		}
	}).AddItem("Exclude all", "", 'x', func() {
		for itx, _ := range content.Files {
			content.Files[itx].Selected = false
			list.SetItemText(itx+preItems, content.Files[itx].Name, "Ignoring")
		}
	})

	for itx, file := range content.Files {
		list = list.AddItem(file.Name, "To download", 'a', func() {
			content.Files[itx].Selected = !content.Files[itx].Selected
			if content.Files[itx].Selected {
				list.SetItemText(itx+preItems, content.Files[itx].Name, "To download")
			} else {
				list.SetItemText(itx+preItems, content.Files[itx].Name, "Ignoring")
			}
		})
	}

	if err := app.SetRoot(list, true).EnableMouse(true).Run(); err != nil {
		panic(err)
	}
}

func getArchiveDetails() archiveorg.ArchiveDetails {
	app := tview.NewApplication()

	form := tview.NewForm().
		AddInputField("Archive name", "", 100, nil, nil).
		AddTextView("", "E.g. for https://archive.org/details/adventures_sherlock_holmes_rg_librivox, enter: adventures_sherlock_holmes_rg_librivox", 100, 2, false, false).
		AddInputField("Cookies", "", 20, nil, nil).
		AddTextView("", "Provide 'cookies' value. Can be seen in cURL command when copied from Firefox network tab", 100, 2, false, false).
		AddInputField("Destination directory", "./", 20, nil, nil).
		AddButton("Continue", func() {
			app.Stop()
		})
	form.SetBorder(true).SetTitle("Enter archive details").SetTitleAlign(tview.AlignLeft)
	if err := app.SetRoot(form, true).EnableMouse(true).EnablePaste(true).Run(); err != nil {
		panic(err)
	}

	return archiveorg.ArchiveDetails{
		Name:                 form.GetFormItem(0).(*tview.InputField).GetText(),
		Cookie:               form.GetFormItem(2).(*tview.InputField).GetText(),
		DestinationDirectory: form.GetFormItem(4).(*tview.InputField).GetText(),
	}
}

func getArchiveContent(archiveDetails *archiveorg.ArchiveDetails) *archiveorg.ArchiveContent {
	// Obtain XML with files in archive
	res, err := archiveDetails.DownloadFile(fmt.Sprintf("%s_files.xml", archiveDetails.Name))
	if err != nil {
		panic(err)
	}
	defer res.Body.Close()

	data, err := io.ReadAll(res.Body)
	if err != nil {
		panic(err)
	}

	var files archiveorg.FilesXml
	err = xml.Unmarshal(data, &files)
	if err != nil {
		panic(err)
	}

	var archiveContent archiveorg.ArchiveContent

	// Remove built-in files and enable remaining files
	for _, file := range files.Files {
		if file.Name == fmt.Sprintf("%s_reviews.xml", archiveDetails.Name) {
			continue
		}

		archiveContent.Files = append(archiveContent.Files, archiveorg.ArchiveFile{
			Name:     file.Name,
			Selected: true,
		})
	}

	return &archiveContent
}
